require "dotenv/parser"
require "dotenv/substitutions/variable"
require "dotenv/substitutions/command" if RUBY_VERSION > "1.8.7"

module Dotenv
  module Init
    class CommentAwareParser < Dotenv::Parser
      @substitutions =
        [Dotenv::Substitutions::Variable, Dotenv::Substitutions::Command]

      LINE = /
        \A
        (?:                        # optional variable setting
          (?:export\s+)?           #   optional export
          ([\w\.]+)                #   key
          (?:\s*(=)\s*|:\s+?)      #   separator
          (                        #   optional value
            '(?:\'|[^'])*'         #     single quoted value
            |                      #     or
            "(?:\"|[^"])*"         #     double quoted value
            |                      #     or
            [^#\n\s]+              #     unquoted value
          )?                       #
        )?                         #
        (?:\s*\#\s*(.*))?          # optional comment
        \z
      /x # '

      COMMENT_OR_BLANK_LINE = /\A\s*(?:#.*)?\z/

      def initialize(*args)
        @line_number = 0
        @comments = []
        super
      end

      def call
        @string.split(/[\n\r]/).each do |line|
          parse_line(line)
          @line_number += 1
        end
        @hash
      end

      private

      def parse_line(line)
        if match = validate_line!(line)
          key, _, value, comment = match.captures
          @comments << comment

          if key
            @hash[key] = {
              default: parse_value(value),
              comments: flush_comments,
              line_number: @line_number,
              offset: offset_for(match, value),
            }
          end
        end
      end

      def offset_for(match, value)
        return [match.offset(2)[1]] * 2 unless value

        offset = match.offset(3)

        if value.start_with?("'", '"')
          [offset[0] + 1, offset[1] - 1]
        else
          offset
        end
      end

      def validate_line!(line)
        if match = line.match(LINE)
          return match
        elsif exports_nothing?(line)
          raise FormatError, "Line #{line.inspect} has an unset variable"
        elsif line !~ COMMENT_OR_BLANK_LINE
          raise FormatError, "Line #{line.inspect} doesn't match format"
        end
      end

      def parse_value(value)
        super(value || "")
      end

      def flush_comments
        comments = @comments
        @comments = []
        comments.compact.map(&:chomp).join(" ")
      end

      def exports_nothing?(line)
        line.split.first == "export" && variable_not_set?(line)
      end
    end
  end
end
